# Copyright (C) 2009-2010  W. Trevor King <wking@drexel.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The author may be contacted at <wking@drexel.edu> on the Internet, or
# write to Trevor King, Drexel University, Physics Dept., 3141 Chestnut St.,
# Philadelphia PA 19104, USA.

"""Histogram generation and comparison.
"""

import numpy


class Histogram (object):
    """A histogram with a flexible comparison method, `residual()`.

    >>> h = Histogram()
    """
    def from_data(self, data, bin_edges):
        """Initialize from `data`.

        All bins should be of equal width (so we can calculate which
        bin a data point belongs to).

        `data` should be a numpy array.
        """
        self.bin_edges = bin_edges
        bin_width = self.bin_edges[1] - self.bin_edges[0]

        bin_is = numpy.floor((data - self.bin_edges[0])/bin_width)
        self.counts = []
        for i in range(len(self.bin_edges)-1):
            self.counts.append(sum(bin_is == i))
        self.total = float(len(data)) # some data might be outside the bins
        self.mean = data.mean()
        self.std_dev = data.std()
        self.normalize()

    def from_stream(self, stream):
        """Initialize from `stream`.

        File format:

            # comment and blank lines ignored
            <bin_edge><whitespace><count>
            ...

        For example:


        `<bin_edge>` should mark the left-hand side of the bin, and
        all bins should be of equal width (so we know where the last
        one ends).

        >>> import StringIO
        >>> h = Histogram()
        >>> h.from_stream(StringIO.StringIO('''# Force (N)\tUnfolding events
        ... 150e-12\t10
        ... 200e-12\t40
        ... 250e-12\t5
        ... '''))
        >>> h.total
        55.0
        >>> h.counts
        [10.0, 40.0, 5.0]
        >>> h.bin_edges  # doctest: +ELLIPSIS
        [1.5e-10, 2.000...e-10, 2.500...e-10, 3e-10]
        >>> h.probabilities  # doctest: +ELLIPSIS
        [0.181..., 0.727..., 0.0909...]
        """
        self.bin_edges = []
        self.counts = []
        for line in stream.readlines():
            line = line.strip()
            if len(line) == 0 or line[0] == "#":
                continue # ignore blank lines and comments
            bin_edge,count = line.split()
            self.bin_edges.append(float(bin_edge))
            self.counts.append(float(count))
        bin_width = self.bin_edges[1] - self.bin_edges[0]
        self.bin_edges.append(self.bin_edges[-1]+bin_width)
        self.total = float(sum(self.counts))
        self.mean = 0
        for bin,count in zip(self.bin_edges, self.counts):
            bin += bin_width / 2.0
            self.mean += bin * count
        self.mean /=  float(self.total)
        variance = 0
        for bin,count in zip(self.bin_edges, self.counts):
            bin += bin_width / 2.0
            variance += (bin - self.mean)**2 * count
        self.std_dev = numpy.sqrt(variance)
        self.normalize()

    def normalize(self):
        self.total = float(self.total)
        self.probabilities = [count/self.total for count in self.counts]

    def mean_residual(self, other):
        return abs(other.mean - self.mean)

    def std_dev_residual(self, other):
        return abs(other.std_dev - self.std_dev)

    def chi_squared_residual(self, other):
        assert self.bin_edges == other.bin_edges
        residual = 0
        for probA,probB in zip(self.probabilities, other.probabilities):
            residual += (probA-probB)**2 / probB
        return residual

    def jensen_shannon_residual(self, other):
        assert self.bin_edges == other.bin_edges
        def d_KL_term(p,q):
            """
            Kullback-Leibler divergence for a single bin, with the
            exception that we return 0 if q==0, rather than
            exploding like d_KL should.  We can do this because
            the way d_KL_term is used in the Jensen-Shannon
            divergence, if q (really m) == 0, then p also == 0,
            and the Jensen-Shannon divergence for the entire term
            should be zero.
            """
            if p==0 or q==0 or p==q:
                return 0.0
            return p * numpy.log2(p/q)
        residual = 0
        for probA,probB in zip(self.probabilities, other.probabilities):
            m = (probA+probB) / 2.0
            residual += 0.5*(d_KL_term(probA,m) + d_KL_term(probB,m))
        return residual

    def _method_to_type(self, name):
        return name[:-len('_residual')].replace('_', '-')

    def _type_to_method(self, name):
        return '%s_residual' % name.replace('-', '_')

    def types(self):
        """Return a list of supported residual types.
        """
        return sorted([self._method_to_type(x)
                       for x in dir(self) if x.endswith('_residual')])

    def residual(self, other, type='jensen-shannon'):
        """Compare this histogram with `other`.

        Supported comparison `type`\s may be found with `types()`:

        >>> h = Histogram()
        >>> h.types()
        ['chi-squared', 'jensen-shannon', 'mean', 'std-dev']

        Selecting an invalid `type` raises an exception.

        >>> h.residual(other=None, type='invalid-type')
        Traceback (most recent call last):
          ...
        AttributeError: 'Histogram' object has no attribute 'invalid_type_residual'

        For residual types where there is a convention, this histogram
        is treated as the experimental histogram and the other
        histogram is treated as the theoretical one.
        """
        r_method = getattr(self, self._type_to_method(type))
        return r_method(other)
